import os
import contextlib

import pytest

from .. import command_line
from .. import yamale_error

dir_path = os.path.dirname(os.path.realpath(__file__))

parsers = ["pyyaml", "PyYAML", "ruamel"]


@contextlib.contextmanager
def scoped_change_dir(new_dir):
    cwd = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(cwd)


@pytest.mark.parametrize("parser", parsers)
def test_bad_yaml(parser):
    with pytest.raises(ValueError) as e:
        command_line._router(
            "yamale/tests/command_line_fixtures/yamls/bad.yaml",
            "schema.yaml",
            1,
            parser,
        )
    assert "map.bad: '12.5' is not a str." in e.value.message


@pytest.mark.parametrize("parser", parsers)
def test_required_keys_yaml(parser):
    with pytest.raises(ValueError) as e:
        command_line._router(
            "yamale/tests/command_line_fixtures/yamls/required_keys_bad.yaml",
            "required_keys_schema.yaml",
            1,
            parser,
        )
    assert "map.key: Required field missing" in e.value.message


@pytest.mark.parametrize("parser", parsers)
def test_good_yaml(parser):
    command_line._router("yamale/tests/command_line_fixtures/yamls/good.yaml", "schema.yaml", 1, parser)


@pytest.mark.parametrize("parser", parsers)
def test_good_relative_yaml(parser):
    command_line._router(
        "yamale/tests/command_line_fixtures/yamls/good.yaml",
        "../schema_dir/external.yaml",
        1,
        parser,
    )


@pytest.mark.parametrize("parser", parsers)
def test_good_relative_schema_in_subfolder(parser):
    with scoped_change_dir("yamale/tests/command_line_fixtures/schema_dir"):
        command_line._router("../yamls/good.yaml", "external.yaml", 1, parser)


@pytest.mark.parametrize("parser", parsers)
def test_external_glob_schema(parser):
    command_line._router(
        "yamale/tests/command_line_fixtures/yamls/good.yaml",
        os.path.join(dir_path, "command_line_fixtures/schema_dir/ex*.yaml"),
        1,
        parser,
    )


def test_empty_schema_file():
    with pytest.raises(ValueError, match="is an empty file!"):
        command_line._router(
            "yamale/tests/command_line_fixtures/empty_schema/data.yaml",
            "empty_schema.yaml",
            1,
            "PyYAML",
        )


def test_external_schema():
    command_line._router(
        "yamale/tests/command_line_fixtures/yamls/good.yaml",
        os.path.join(dir_path, "command_line_fixtures/schema_dir/external.yaml"),
        1,
        "PyYAML",
    )


def test_bad_dir():
    with pytest.raises(ValueError):
        command_line._router("yamale/tests/command_line_fixtures/yamls", "schema.yaml", 4, "PyYAML")


def test_bad_strict():
    with pytest.raises(ValueError) as e:
        command_line._router(
            "yamale/tests/command_line_fixtures/yamls/required_keys_extra_element.yaml",
            "required_keys_schema.yaml",
            4,
            "PyYAML",
            strict=True,
        )
    assert "map.key2: Unexpected element" in e.value.message


def test_bad_issue_54():
    with pytest.raises(yamale_error.YamaleError) as e:
        command_line._router(
            "yamale/tests/fixtures/nested_issue_54.yaml",
            "nested.yaml",
            4,
            "PyYAML",
            strict=True,
        )
    assert "string: Required field missing" in e.value.message
    assert "number: Required field missing" in e.value.message
    assert "integer: Required field missing" in e.value.message
    assert "boolean: Required field missing" in e.value.message
    assert "date: Required field missing" in e.value.message
    assert "datetime: Required field missing" in e.value.message
    assert "nest: Required field missing" in e.value.message
    assert "list: Required field missing" in e.value.message


def test_nested_schema_issue_69():
    command_line._router("yamale/tests/command_line_fixtures/nestedYaml", "schema.yaml", 1, "PyYAML")



import re
import os
from yamale import YamaleTestCase
from yamale.validators import DefaultValidators, Validator


data_folder = os.path.dirname(os.path.realpath(__file__))


class TestAllYaml(YamaleTestCase):
    base_dir = data_folder
    schema = "meta_test_fixtures/schema.yaml"
    yaml = "meta_test_fixtures/data1.yaml"

    def runTest(self):
        self.assertTrue(self.validate())


class TestBadYaml(YamaleTestCase):
    base_dir = data_folder
    schema = "meta_test_fixtures/schema_bad.yaml"
    yaml = "meta_test_fixtures/data*.yaml"

    def runTest(self):
        self.assertRaises(ValueError, self.validate)


class TestMapYaml(YamaleTestCase):
    base_dir = data_folder
    schema = "meta_test_fixtures/schema.yaml"
    yaml = [
        "meta_test_fixtures/data1.yaml",
        "meta_test_fixtures/some_data.yaml",
        # Make sure  schema doesn't validate itself
        "meta_test_fixtures/schema.yaml",
    ]

    def runTest(self):
        self.assertTrue(self.validate())


# class TestListYaml(YamaleTestCase):
#     base_dir = data_folder
#     schema = 'meta_test_fixtures/schema_include_list.yaml'
#     yaml = ['meta_test_fixtures/data_include_list.yaml']

#     def runTest(self):
#         self.assertTrue(self.validate())


class Card(Validator):
    """Custom validator for testing purpose"""

    tag = "card"
    card_regex = re.compile(r"^(10|[2-9JQKA])[SHDC]$")

    def _is_valid(self, value):
        return re.match(self.card_regex, value)


class TestCustomValidator(YamaleTestCase):
    base_dir = data_folder
    schema = "meta_test_fixtures/schema_custom.yaml"
    yaml = "meta_test_fixtures/data_custom.yaml"

    def runTest(self):
        validators = DefaultValidators.copy()
        validators["card"] = Card
        self.assertTrue(self.validate(validators))


class TestCustomValidatorWithIncludes(YamaleTestCase):
    base_dir = data_folder
    schema = "meta_test_fixtures/schema_custom_with_include.yaml"
    yaml = "meta_test_fixtures/data_custom_with_include.yaml"

    def runTest(self):
        validators = DefaultValidators.copy()
        validators["card"] = Card
        self.assertTrue(self.validate(validators))


class TestBadRequiredYaml(YamaleTestCase):
    base_dir = data_folder
    schema = "meta_test_fixtures/schema_required_bad.yaml"
    yaml = "meta_test_fixtures/data_required_bad.yaml"

    def runTest(self):
        self.assertRaises(ValueError, self.validate)


class TestGoodRequiredYaml(YamaleTestCase):
    base_dir = data_folder
    schema = "meta_test_fixtures/schema_required_good.yaml"
    yaml = "meta_test_fixtures/data_required_good.yaml"

    def runTest(self):
        self.assertTrue(self.validate())



from .datapath import DataPath
from .validationresults import ValidationResult
from .. import syntax, util
from .. import validators as val


class Schema(object):
    """
    Makes a Schema object from a schema dict.
    Still acts like a dict.
    """

    def __init__(self, schema_dict, name="", validators=None, includes=None):
        self.validators = validators or val.DefaultValidators
        self.dict = schema_dict
        self.name = name
        self._schema = self._process_schema(DataPath(), schema_dict, self.validators)
        # if this schema is included it shares the includes with the top level
        # schema
        self.includes = {} if includes is None else includes

    def add_include(self, type_dict):
        for include_name, custom_type in type_dict.items():
            t = Schema(custom_type, name=include_name, validators=self.validators, includes=self.includes)
            self.includes[include_name] = t

    def _process_schema(self, path, schema_data, validators):
        """
        Go through a schema and construct validators.
        """
        if util.is_map(schema_data) or util.is_list(schema_data):
            for key, data in util.get_iter(schema_data):
                schema_data[key] = self._process_schema(path + DataPath(key), data, validators)
        else:
            schema_data = self._parse_schema_item(path, schema_data, validators)
        return schema_data

    def _parse_schema_item(self, path, expression, validators):
        try:
            return syntax.parse(expression, validators)
        except SyntaxError as e:
            # Tack on some more context and rethrow.
            error = str(e) + " at node '%s'" % str(path)
            raise SyntaxError(error)

    def validate(self, data, data_name, strict):
        path = DataPath()
        errors = self._validate(self._schema, data, path, strict)
        return ValidationResult(data_name, self.name, errors)

    def _validate_item(self, validator, data, path, strict, key):
        """
        Fetch item from data at the position key and validate with validator.

        Returns an array of errors.
        """
        errors = []
        path = path + DataPath(key)
        try:  # Pull value out of data. Data can be a map or a list/sequence
            data_item = data[key]
        except (KeyError, IndexError):  # Oops, that field didn't exist.
            # Optional? Who cares.
            if isinstance(validator, val.Validator) and validator.is_optional:
                return errors
            # SHUT DOWN EVERYTHING
            errors.append("%s: Required field missing" % path)
            return errors

        return self._validate(validator, data_item, path, strict)

    def _validate(self, validator, data, path, strict):
        """
        Validate data with validator.
        Special handling of non-primitive validators.

        Returns an array of errors.
        """

        if util.is_list(validator) or util.is_map(validator):
            return self._validate_static_map_list(validator, data, path, strict)

        errors = []
        # Optional field with optional value? Who cares.
        if data is None and validator.is_optional and validator.can_be_none:
            return errors

        errors += self._validate_primitive(validator, data, path)

        if errors:
            return errors

        if isinstance(validator, val.Include):
            errors += self._validate_include(validator, data, path, strict)

        elif isinstance(validator, (val.Map, val.List)):
            errors += self._validate_map_list(validator, data, path, strict)

        elif isinstance(validator, val.Any):
            errors += self._validate_any(validator, data, path, strict)

        elif isinstance(validator, val.Subset):
            errors += self._validate_subset(validator, data, path, strict)

        return errors

    def _validate_static_map_list(self, validator, data, path, strict):
        if util.is_map(validator) and not util.is_map(data):
            return ["%s : '%s' is not a map" % (path, data)]

        if util.is_list(validator) and not util.is_list(data):
            return ["%s : '%s' is not a list" % (path, data)]

        errors = []

        if strict:
            data_keys = set(util.get_keys(data))
            validator_keys = set(util.get_keys(validator))
            for key in data_keys - validator_keys:
                error_path = path + DataPath(key)
                errors += ["%s: Unexpected element" % error_path]

        for key, sub_validator in util.get_iter(validator):
            errors += self._validate_item(sub_validator, data, path, strict, key)
        return errors

    def _validate_map_list(self, validator, data, path, strict):
        errors = []

        if not validator.validators:
            return errors  # No validators, user just wanted a map.

        for key in util.get_keys(data):
            sub_errors = []
            for v in validator.validators:
                err = self._validate_item(v, data, path, strict, key)
                if err:
                    sub_errors.append(err)

            if len(sub_errors) == len(validator.validators):
                # All validators failed, add to errors
                for err in sub_errors:
                    errors += err

        return errors

    def _validate_include(self, validator, data, path, strict):
        include_schema = self.includes.get(validator.include_name)
        if not include_schema:
            return [("Include '%s' has not been defined." % validator.include_name)]
        strict = strict if validator.strict is None else validator.strict
        return include_schema._validate(include_schema._schema, data, path, strict)

    def _validate_any(self, validator, data, path, strict):
        if not validator.validators:
            return []

        errors = []

        sub_errors = []
        for v in validator.validators:
            err = self._validate(v, data, path, strict)
            if err:
                sub_errors.append(err)

        if len(sub_errors) == len(validator.validators):
            # All validators failed, add to errors
            for err in sub_errors:
                errors += err

        return errors

    def _validate_subset(self, validator, data, path, strict):
        def _internal_validate(internal_data):
            sub_errors = []
            for val in validator.validators:
                err = self._validate(val, internal_data, path, strict)
                if not err:
                    break
                sub_errors += err
            else:
                return sub_errors
            return []

        if not validator.validators:
            return []

        errors = []
        if util.is_map(data):
            for k, v in data.items():
                errors += _internal_validate({k: v})
        elif util.is_list(data):
            for k in data:
                errors += _internal_validate(k)
        else:
            errors += _internal_validate(data)
        return errors

    def _validate_primitive(self, validator, data, path):
        errors = validator.validate(data)

        for i, error in enumerate(errors):
            errors[i] = ("%s: " % path) + error

        return errors



import datetime
from yamale import validators as val


def test_length_min():
    v = val.String(min=2)
    assert v.is_valid("abcd")
    assert v.is_valid("ab")
    assert not v.is_valid("a")


def test_length_max():
    v = val.String(max=3)
    assert v.is_valid("abc")
    assert v.is_valid("ab")
    assert not v.is_valid("abcd")


def test_number_max():
    v = val.Number(min=0.5)
    assert v.is_valid(4)
    assert v.is_valid(0.5)
    assert not v.is_valid(0.1)


def test_number_min():
    v = val.Integer(max=10)
    assert v.is_valid(4)
    assert v.is_valid(10)
    assert not v.is_valid(11)


def test_timestamp_min():
    v = val.Timestamp(min=datetime.datetime(2010, 1, 1))
    assert v.is_valid(datetime.datetime(2010, 1, 1))
    assert v.is_valid(datetime.datetime(2011, 2, 2))
    assert not v.is_valid(datetime.datetime(2009, 12, 31))


def test_timestamp_max():
    v = val.Timestamp(max=datetime.datetime(2010, 1, 1))
    assert v.is_valid(datetime.datetime(2010, 1, 1))
    assert v.is_valid(datetime.datetime(2009, 2, 2))
    assert not v.is_valid(datetime.datetime(2010, 2, 2))


def test_day_min():
    v = val.Day(min=datetime.date(2010, 1, 1))
    assert v.is_valid(datetime.date(2010, 1, 1))
    assert v.is_valid(datetime.date(2011, 2, 2))
    assert not v.is_valid(datetime.date(2009, 12, 31))


def test_day_max():
    v = val.Day(max=datetime.date(2010, 1, 1))
    assert v.is_valid(datetime.date(2010, 1, 1))
    assert v.is_valid(datetime.date(2009, 2, 2))
    assert not v.is_valid(datetime.date(2010, 2, 2))


def test_str_equals():
    v = val.String(equals="abcd")
    assert v.is_valid("abcd")
    assert not v.is_valid("abcde")
    assert not v.is_valid("c")


def test_str_equals_ignore_case():
    v = val.String(equals="abcd", ignore_case=True)
    assert v.is_valid("abCd")
    assert not v.is_valid("abcde")
    assert not v.is_valid("C")


def test_str_starts_with():
    v = val.String(starts_with="abc")
    assert v.is_valid("abcd")
    assert not v.is_valid("bcd")
    assert not v.is_valid("c")


def test_str_starts_with_ignore_case():
    v = val.String(starts_with="abC", ignore_case=True)
    assert v.is_valid("abCde")
    assert v.is_valid("abcde")
    assert not v.is_valid("bcd")
    assert not v.is_valid("C")


def test_str_ends_with():
    v = val.String(ends_with="abcd")
    assert v.is_valid("abcd")
    assert not v.is_valid("abcde")
    assert not v.is_valid("c")


def test_str_ends_with_ignore_case():
    v = val.String(ends_with="abC", ignore_case=True)
    assert v.is_valid("xyzabC")
    assert v.is_valid("xyzabc")
    assert not v.is_valid("cde")
    assert not v.is_valid("C")


def test_str_matches():
    v = val.String(matches=r"^(abc)\1?de$")
    assert v.is_valid("abcabcde")
    assert not v.is_valid("abcabcabcde")
    assert not v.is_valid("\12")

    v = val.String(matches=r"[a-z0-9]{3,}s\s$", ignore_case=True)
    assert v.is_valid("b33S\v")
    assert v.is_valid("B33s\t")
    assert not v.is_valid(" b33s ")
    assert not v.is_valid("b33s  ")

    v = val.String(matches=r"A.+\d$", ignore_case=False, multiline=True)
    assert v.is_valid("A_-3\n\n")
    assert not v.is_valid("a!!!!!5\n\n")

    v = val.String(matches=r".*^Ye.*s\.", ignore_case=True, multiline=True, dotall=True)
    assert v.is_valid("YEeeEEEEeeeeS.")
    assert v.is_valid("What?\nYes!\nBEES.\nOK.")
    assert not v.is_valid("YES-TA-TOES?")
    assert not v.is_valid("\n\nYaes.")


def test_char_exclude():
    v = val.String(exclude="abcd")
    assert v.is_valid("efg")
    assert not v.is_valid("abc")
    assert not v.is_valid("c")


def test_char_exclude_igonre_case():
    v = val.String(exclude="abcd", ignore_case=True)
    assert v.is_valid("efg")
    assert v.is_valid("Efg")
    assert not v.is_valid("abc")
    assert not v.is_valid("Def")
    assert not v.is_valid("c")


def test_ip4():
    v = val.Ip(version=4)
    assert v.is_valid("192.168.1.1")
    assert v.is_valid("192.168.1.255")
    assert v.is_valid("192.168.3.1/24")
    assert not v.is_valid("2001:db8::")
    assert not v.is_valid("2001:db8::/64")


def test_ip6():
    v = val.Ip(version=6)
    assert not v.is_valid("192.168.1.1")
    assert not v.is_valid("192.168.1.255")
    assert not v.is_valid("192.168.3.1/24")
    assert v.is_valid("2001:db8::")
    assert v.is_valid("2001:db8::/64")



# ABCs for containers were moved to their own module
try:
    from collections.abc import Mapping, Set, Sequence
except ImportError:
    from collections import Mapping, Set, Sequence


# Python 3 has no basestring, lets test it.
try:
    basestring  # attempt to evaluate basestring

    def isstr(s):
        return isinstance(s, basestring)

    def to_unicode(s):
        return unicode(s)

except NameError:

    def isstr(s):
        return isinstance(s, str)

    def to_unicode(s):
        return s


def is_list(obj):
    return isinstance(obj, Sequence) and not isstr(obj)


def is_map(obj):
    return isinstance(obj, Mapping)


def get_keys(obj):
    if is_map(obj):
        return obj.keys()
    elif is_list(obj):
        return range(len(obj))


def get_iter(iterable):
    if isinstance(iterable, Mapping):
        return iterable.items()
    else:
        return enumerate(iterable)


def get_subclasses(cls, _subclasses_yielded=None):
    """
    Generator recursively yielding all subclasses of the passed class (in
    depth-first order).

    Parameters
    ----------
    cls : type
        Class to find all subclasses of.
    _subclasses_yielded : set
        Private parameter intended to be passed only by recursive invocations of
        this function, containing all previously yielded classes.
    """

    if _subclasses_yielded is None:
        _subclasses_yielded = set()

    # If the passed class is old- rather than new-style, raise an exception.
    if not hasattr(cls, "__subclasses__"):
        raise TypeError('Old-style class "%s" unsupported.' % cls.__name__)

    # For each direct subclass of this class
    for subclass in cls.__subclasses__():
        # If this subclass has already been yielded, skip to the next.
        if subclass in _subclasses_yielded:
            continue

        # Yield this subclass and record having done so before recursing.
        yield subclass
        _subclasses_yielded.add(subclass)

        # Yield all direct subclasses of this class as well.
        for subclass_subclass in get_subclasses(subclass, _subclasses_yielded):
            yield subclass_subclass



import re
from datetime import date, datetime
import ipaddress
from .base import Validator
from . import constraints as con
from .. import util

# ABCs for containers were moved to their own module
try:
    from collections.abc import Sequence, Mapping
except ImportError:
    from collections import Sequence, Mapping


class String(Validator):
    """String validator"""

    tag = "str"
    constraints = [
        con.LengthMin,
        con.LengthMax,
        con.CharacterExclude,
        con.StringEquals,
        con.StringStartsWith,
        con.StringEndsWith,
        con.StringMatches,
    ]

    def _is_valid(self, value):
        return util.isstr(value)


class Number(Validator):
    """Number/float validator"""

    value_type = float
    tag = "num"
    constraints = [con.Min, con.Max]

    def _is_valid(self, value):
        return isinstance(value, (int, float)) and not isinstance(value, bool)


class Integer(Validator):
    """Integer validator"""

    value_type = int
    tag = "int"
    constraints = [con.Min, con.Max]

    def _is_valid(self, value):
        return isinstance(value, int) and not isinstance(value, bool)


class Boolean(Validator):
    """Boolean validator"""

    tag = "bool"

    def _is_valid(self, value):
        return isinstance(value, bool)


class Enum(Validator):
    """Enum validator"""

    tag = "enum"

    def __init__(self, *args, **kwargs):
        super(Enum, self).__init__(*args, **kwargs)
        self.enums = args

    def _is_valid(self, value):
        return value in self.enums

    def fail(self, value):
        return "'%s' not in %s" % (value, self.enums)


class Day(Validator):
    """Day validator. Format: YYYY-MM-DD"""

    value_type = date
    tag = "day"
    constraints = [con.Min, con.Max]

    def _is_valid(self, value):
        return isinstance(value, date)


class Timestamp(Validator):
    """Timestamp validator. Format: YYYY-MM-DD HH:MM:SS"""

    value_type = datetime
    tag = "timestamp"
    constraints = [con.Min, con.Max]

    def _is_valid(self, value):
        return isinstance(value, datetime)


class Map(Validator):
    """Map and dict validator"""

    tag = "map"
    constraints = [con.LengthMin, con.LengthMax, con.Key]

    def __init__(self, *args, **kwargs):
        super(Map, self).__init__(*args, **kwargs)
        self.validators = [val for val in args if isinstance(val, Validator)]

    def _is_valid(self, value):
        return isinstance(value, Mapping)


class List(Validator):
    """List validator"""

    tag = "list"
    constraints = [con.LengthMin, con.LengthMax]

    def __init__(self, *args, **kwargs):
        super(List, self).__init__(*args, **kwargs)
        self.validators = [val for val in args if isinstance(val, Validator)]

    def _is_valid(self, value):
        return isinstance(value, Sequence) and not util.isstr(value)


class Include(Validator):
    """Include validator"""

    tag = "include"

    def __init__(self, *args, **kwargs):
        self.include_name = args[0]
        self.strict = kwargs.pop("strict", None)
        super(Include, self).__init__(*args, **kwargs)

    def _is_valid(self, value):
        return True

    def get_name(self):
        return self.include_name


class Any(Validator):
    """Any of several types validator"""

    tag = "any"

    def __init__(self, *args, **kwargs):
        self.validators = [val for val in args if isinstance(val, Validator)]
        super(Any, self).__init__(*args, **kwargs)

    def _is_valid(self, value):
        return True


class Subset(Validator):
    """Subset of several types validator"""

    tag = "subset"

    def __init__(self, *args, **kwargs):
        super(Subset, self).__init__(*args, **kwargs)
        self._allow_empty_set = bool(kwargs.pop("allow_empty", False))
        self.validators = [val for val in args if isinstance(val, Validator)]
        if not self.validators:
            raise ValueError("'%s' requires at least one validator!" % self.tag)

    def _is_valid(self, value):
        return self.can_be_none or value is not None

    def fail(self, value):
        # Called in case `_is_valid` returns False
        return "'%s' may not be an empty set." % self.get_name()

    @property
    def is_optional(self):
        return self._allow_empty_set

    @property
    def can_be_none(self):
        return self._allow_empty_set


class Null(Validator):
    """Validates null"""

    value_type = None
    tag = "null"

    def _is_valid(self, value):
        return value is None


class Regex(Validator):
    """Regular expression validator"""

    tag = "regex"
    _regex_flags = {"ignore_case": re.I, "multiline": re.M, "dotall": re.S}

    def __init__(self, *args, **kwargs):
        self.regex_name = kwargs.pop("name", None)

        flags = 0
        for k, v in util.get_iter(self._regex_flags):
            flags |= v if kwargs.pop(k, False) else 0

        self.regexes = [re.compile(arg, flags) for arg in args if util.isstr(arg)]
        super(Regex, self).__init__(*args, **kwargs)

    def _is_valid(self, value):
        return util.isstr(value) and any(r.match(value) for r in self.regexes)

    def get_name(self):
        return self.regex_name or self.tag + " match"


class Ip(Validator):
    """IP address validator"""

    tag = "ip"
    constraints = [con.IpVersion]

    def _is_valid(self, value):
        return self.ip_address(value)

    def ip_address(self, value):
        try:
            ipaddress.ip_interface(util.to_unicode(value))
        except ValueError:
            return False
        return True


class Mac(Regex):
    """MAC address validator"""

    tag = "mac"

    def __init__(self, *args, **kwargs):
        super(Mac, self).__init__(*args, **kwargs)
        self.regexes = [
            re.compile("[0-9a-fA-F]{2}([-:]?)[0-9a-fA-F]{2}(\\1[0-9a-fA-F]{2}){4}$"),
            re.compile("[0-9a-fA-F]{4}([-:]?)[0-9a-fA-F]{4}(\\1[0-9a-fA-F]{4})$"),
        ]


DefaultValidators = {}

for v in util.get_subclasses(Validator):
    # Allow validator nodes to contain either tags or actual name
    DefaultValidators[v.tag] = v
    DefaultValidators[v.__name__] = v
